"""
This module contains functionality for reading statistics data from
the PacketLogic system.

Do NOT import this module directly. The L{PLConnection
<packetlogic2.PLConnection>} object should be used, e.g:

    >>> import packetlogic2
    >>> pl = packetlogic2.connect("192.168.1.25", "admin", "pldemo00")
    >>> s = pl.Statistics()

@var __revision__: Revision identifier.
"""

__revision__ = '%K%'

from resource import Resource
from thisversion import PLDB

import plutil

from packetlogic2.exceptions import PLDBUnsupportedInFirmware

_NONEXISTANT_VALUETYPE=-1
_NONEXISTANT_FIELD=-1
_NONEXISTANT_GRAPH_MODE=-1

class Priority(int):	
	priority = property(lambda self: self & 0x7f)
	expandable = property(lambda self: bool(self & 0x80))
	
class Statistics(Resource):
	"""The Statistics resource is used for for reading statistics.

	    @group Value types: VALUETYPE_ROOT, VALUETYPE_AGGREGATION, VALUETYPE_TRAFFICSTATS, VALUETYPE_ZONESTATS, VALUETYPE_CHANNELSTATS, VALUETYPE_NETOBJECT, VALUETYPE_HOST, VALUETYPE_LOCALVHOST, VALUETYPE_REMOTEVHOST, VALUETYPE_SERVOBJECT, VALUETYPE_SERVICE, VALUETYPE_CLIENTASPATH, VALUETYPE_SERVERASPATH, VALUETYPE_VLAN, VALUETYPE_INVLAN, VALUETYPE_OUTVLAN, VALUETYPE_INDSCP, VALUETYPE_OUTDSCP, VALUETYPE_INCHAN, VALUETYPE_OUTCHAN, VALUETYPE_INMPLS, VALUETYPE_OUTMPLS
	    @group Total Fields: TOTALFIELD_BYTES_IN, TOTALFIELD_BYTES_OUT, TOTALFIELD_CONNS, TOTALFIELD_UCONNS, TOTALFIELD_CONNS_IN, TOTALFIELD_CONNS_OUT, TOTALFIELD_QUALITY_IN, TOTALFIELD_QUALITY_OUT, TOTALFIELD_QUALITY_IN_INT, TOTALFIELD_QUALITY_OUT_INT, TOTALFIELD_QUALITY_IN_EXT, TOTALFIELD_QUALITY_OUT_EXT
	    @group Graph Fields: GRAPHFIELD_BPS_IN, GRAPHFIELD_BPS_OUT,GRAPHFIELD_CPS, GRAPHFIELD_UCPS, GRAPHFIELD_CPS_IN, GRAPHFIELD_CPS_OUT, GRAPHFIELD_QUALITY_IN, GRAPHFIELD_QUALITY_OUT, GRAPHFIELD_QUALITY_IN_INT, GRAPHFIELD_QUALITY_OUT_INT, GRAPHFIELD_QUALITY_IN_EXT, GRAPHFIELD_QUALITY_OUT_EXT
	    @group Graph Modes: GRAPH_MODE_SIZE, GRAPH_MODE_SPEED

	    @cvar VALUETYPE_DUMMY: Root values (/ or Statistics Objects)
	    @cvar VALUETYPE_ROOT: Alias for VALUETYPE_DUMMY
	    @cvar VALUETYPE_AGGREGATION: Aggregation value
	    @cvar VALUETYPE_TRAFFICSTATS: Traffic stats value
	    @cvar VALUETYPE_ZONESTATS: Zone Stats value
	    @cvar VALUETYPE_CHANNELSTATS: Channel Stats value

	    @cvar VALUETYPE_NETOBJECT: Netobject
	    @cvar VALUETYPE_HOST: Host
	    @cvar VALUETYPE_LOCALVHOST: Local Vhost
	    @cvar VALUETYPE_REMOTEVHOST: Remote Vhost
	    @cvar VALUETYPE_SERVOBJECT: ServiceObject
	    @cvar VALUETYPE_SERVICE: Service
	    @cvar VALUETYPE_CLIENTASPATH: Client ASpath
	    @cvar VALUETYPE_SERVERASPATH: Server ASpath
	    @cvar VALUETYPE_VLAN: Vlan
	    @cvar VALUETYPE_INVLAN: Inbound VLAN
	    @cvar VALUETYPE_OUTVLAN: Outbound VLAN
	    @cvar VALUETYPE_INDSCP: Inbound DSCP
	    @cvar VALUETYPE_OUTDSCP: Outbound DSCP
	    @cvar VALUETYPE_INCHAN: Inbound Channel
	    @cvar VALUETYPE_OUTCHAN: Outbound Channel
	    @cvar VALUETYPE_INMPLS: Inbound MPLS
	    @cvar VALUETYPE_OUTMPLS: Outbound MPLS

	    @cvar TOTALFIELD_BYTES_IN: Incoming Bytes count 
	    @cvar TOTALFIELD_BYTES_OUT: Outgoing Bytes count
	    @cvar TOTALFIELD_CONNS: Connections 
	    @cvar TOTALFIELD_UCONNS: Unestablished Connections 
	    @cvar TOTALFIELD_CONNS_IN: Incoming Connections
	    @cvar TOTALFIELD_CONNS_OUT: Outgoing Connections
	    @cvar TOTALFIELD_QUALITY_IN: Incoming Quality of Experience
	    @cvar TOTALFIELD_QUALITY_OUT: Outgoing Quality of Experience
	    @cvar TOTALFIELD_QUALITY_IN_INT: Incoming Quality of Experience (Internal)
	    @cvar TOTALFIELD_QUALITY_OUT_INT: Outgoing Quality of Experience (Internal)
	    @cvar TOTALFIELD_QUALITY_IN_EXT: Incoming Quality of Experience (External)
	    @cvar TOTALFIELD_QUALITY_OUT_EXT: Outgoing Quality of Experience (External)

	    @cvar GRAPHFIELD_BPS_IN: Incoming bytes/bps
	    @cvar GRAPHFIELD_BPS_OUT: Outgoing bytes/bps
	    @cvar GRAPHFIELD_CPS: Connections (per sec)
	    @cvar GRAPHFIELD_UCPS: Unestablished Connections (per sec)
	    @cvar GRAPHFIELD_CPS_IN: Incoming Connections (per sec)
	    @cvar GRAPHFIELD_CPS_OUT:Outgoing Connections
	    @cvar GRAPHFIELD_QUALITY_IN: Incoming Quality of Experience
	    @cvar GRAPHFIELD_QUALITY_OUT: Outgoing Quality of Experience
	    @cvar GRAPHFIELD_QUALITY_IN_INT: Incoming Quality of Experience (Internal)
	    @cvar GRAPHFIELD_QUALITY_OUT_INT: Outgoing Quality of Experience (Internal)
	    @cvar GRAPHFIELD_QUALITY_IN_EXT: Incoming Quality of Experience (External)
	    @cvar GRAPHFIELD_QUALITY_OUT_EXT: Outgoing Quality of Experience (External)

	    @cvar GRAPH_MODE_SIZE: Graph result as absolute count
	    @cvar GRAPH_MODE_SPEED: Graph result as rate.
	"""

	VALUETYPE_DUMMY = 0
	VALUETYPE_ROOT = VALUETYPE_DUMMY
	VALUETYPE_AGGREGATION = _NONEXISTANT_VALUETYPE
	VALUETYPE_TRAFFICSTATS =  _NONEXISTANT_VALUETYPE
	VALUETYPE_ZONESTATS =  _NONEXISTANT_VALUETYPE
	VALUETYPE_CHANNELSTATS = _NONEXISTANT_VALUETYPE

	VALUETYPE_NETOBJECT = _NONEXISTANT_VALUETYPE
	VALUETYPE_HOST = _NONEXISTANT_VALUETYPE
	VALUETYPE_LOCALVHOST = _NONEXISTANT_VALUETYPE
	VALUETYPE_REMOTEVHOST = _NONEXISTANT_VALUETYPE
	VALUETYPE_SERVOBJECT = _NONEXISTANT_VALUETYPE
	VALUETYPE_SERVICE = _NONEXISTANT_VALUETYPE
	VALUETYPE_INTASPATH = _NONEXISTANT_VALUETYPE
	VALUETYPE_EXTASPATH = _NONEXISTANT_VALUETYPE
	VALUETYPE_VLAN = _NONEXISTANT_VALUETYPE
	VALUETYPE_CLIENTASPATH = _NONEXISTANT_VALUETYPE
	VALUETYPE_SERVERASPATH = _NONEXISTANT_VALUETYPE
	VALUETYPE_INVLAN = _NONEXISTANT_VALUETYPE
	VALUETYPE_OUTVLAN = _NONEXISTANT_VALUETYPE
	VALUETYPE_INDSCP = _NONEXISTANT_VALUETYPE
	VALUETYPE_OUTDSCP = _NONEXISTANT_VALUETYPE
	VALUETYPE_INCHAN = _NONEXISTANT_VALUETYPE
	VALUETYPE_OUTCHAN = _NONEXISTANT_VALUETYPE
	VALUETYPE_INMPLS = _NONEXISTANT_VALUETYPE
	VALUETYPE_OUTMPLS = _NONEXISTANT_VALUETYPE

	TOTALFIELD_BYTES_IN = _NONEXISTANT_FIELD
	TOTALFIELD_BYTES_OUT = _NONEXISTANT_FIELD
	TOTALFIELD_CONNS = _NONEXISTANT_FIELD
	TOTALFIELD_UCONNS = _NONEXISTANT_FIELD
	TOTALFIELD_CONNS_IN = _NONEXISTANT_FIELD
	TOTALFIELD_CONNS_OUT = _NONEXISTANT_FIELD
	TOTALFIELD_QUALITY_IN = _NONEXISTANT_FIELD
	TOTALFIELD_QUALITY_OUT = _NONEXISTANT_FIELD
	TOTALFIELD_QUALITY_IN_INT = _NONEXISTANT_FIELD
	TOTALFIELD_QUALITY_OUT_INT = _NONEXISTANT_FIELD
	TOTALFIELD_QUALITY_IN_EXT = _NONEXISTANT_FIELD
	TOTALFIELD_QUALITY_OUT_EXT = _NONEXISTANT_FIELD

	GRAPHFIELD_BPS_IN = _NONEXISTANT_FIELD
	GRAPHFIELD_BPS_OUT = _NONEXISTANT_FIELD
	GRAPHFIELD_CPS = _NONEXISTANT_FIELD
	GRAPHFIELD_UCPS = _NONEXISTANT_FIELD
	GRAPHFIELD_CPS_IN = _NONEXISTANT_FIELD
	GRAPHFIELD_CPS_OUT = _NONEXISTANT_FIELD
	GRAPHFIELD_QUALITY_IN = _NONEXISTANT_FIELD
	GRAPHFIELD_QUALITY_OUT = _NONEXISTANT_FIELD
	GRAPHFIELD_QUALITY_IN_INT = _NONEXISTANT_FIELD
	GRAPHFIELD_QUALITY_OUT_INT = _NONEXISTANT_FIELD
	GRAPHFIELD_QUALITY_IN_EXT = _NONEXISTANT_FIELD
	GRAPHFIELD_QUALITY_OUT_EXT = _NONEXISTANT_FIELD

	GRAPH_MODE_SIZE = _NONEXISTANT_GRAPH_MODE
	GRAPH_MODE_SPEED = 1

	def _resource(self):
		return PLDB.RESOURCE_STAT

	def date_list(self):
		"""
		List the statistics database of available statistics data.

			>>> x.date_list()
			['2005-12-01', '2005-12-02', '2005-12-03', '2005-12-04', '2005-12-05', '2005-12-06', '2005-12-07', '2005-12-08', '2005-12-09']

		@rtype: C{list} of C{str}
		@returns: a sorted list with dates that have statistics data

		"""
		lst = plutil._do(self, self.session.stat_date_list(), plutil._SimpleListCB(self, self.session.stat_date_list_handle))
		if lst is None:
			return lst

		# Some sanity check that this seems to be a date and not e.g lost+found
		dates = [d for d in lst if len(d) == 10 and d[4] + d[7] == '--' and (d[0:4]+d[5:7]+d[-2:]).isdigit()]
		dates.sort()
		
		return dates

	def fields_list(self):
		"""
		List available fields.

			>>> s.fields_list()
			{'graphs': [{'desc': 'Incoming bps', 'id': 1, 'type': 1},
				    {'desc': 'Outgoing bps', 'id': 2, 'type': 1},
				    {'desc': 'CPS', 'id': 4, 'type': 2},
				    {'desc': 'Unestablished CPS', 'id': 8, 'type': 2},
				    {'desc': 'Incoming CPS', 'id': 16, 'type': 2},
				    {'desc': 'Outgoing CPS', 'id': 32, 'type': 2},
				    {'desc': 'Unestablished Incoming CPS', 'id': 64, 'type': 2},
				    {'desc': 'Unestablished Outgoing CPS', 'id': 128, 'type': 2},
				    {'desc': 'Incoming concurrrent connections', 'id': 256, 'type': 2},
				    {'desc': 'Outgoing concurrent connections', 'id': 512, 'type': 2},
				    {'desc': 'Incoming Quality of Experience', 'id': 1024, 'type': 3},
				    {'desc': 'Outgoing Quality of Experience', 'id': 2048, 'type': 3}],
			 'totals': [{'desc': 'Incoming Bytes', 'id': 1, 'type': 1},
				    {'desc': 'Outgoing Bytes', 'id': 2, 'type': 1},
				    {'desc': 'Connections', 'id': 4, 'type': 2},
				    {'desc': 'Unestablished Connections', 'id': 8, 'type': 2},
				    {'desc': 'Incoming Connections', 'id': 16, 'type': 2},
				    {'desc': 'Outgoing Connections', 'id': 32, 'type': 2},
				    {'desc': 'Incoming Unestablished Connections', 'id': 64, 'type': 2},
				    {'desc': 'Outgoing Unestablished Connections', 'id': 128, 'type': 2},
				    {'desc': 'Incoming Quality of Experience', 'id': 256, 'type': 3},
				    {'desc': 'Outgoing Quality of Experience', 'id': 512, 'type': 3}]}

		@returns: Dict with 'graphs' and 'totals'
		sub-dicts. The sub-dicts contain the keys 'desc', 'id'
		and 'type'. Type corresponds to the pathtype used in
		list/graph calls.
		
		@rtype: C{dict} of C{dict}
		"""
		return {'graphs': [{'desc': 'in', 'id': 1, 'type': 0},
				   {'desc': 'out', 'id': 2, 'type': 0}],
			'totals': [{'desc': 'in', 'id': 1, 'type': 0},
				   {'desc': 'out', 'id': 2, 'type': 0}]}
		

	def list(self, start, end, path, pathtype=0):
		"""
		List stats nodes between two dates

			>>> s.list("2005-12-01", "2005-12-02", "/")
			[{'name': 'HostperNO', 
				'values': {'bytes out': 4900070621184.0,
					'connections': 351976704.0, 
					'outbound Quality of Experience': 99.520118713378906, 
					'inbound Quality of Experience': 99.921859741210938, 
					'bytes in': 21536893304832.0}, 
				'flags': 3,'type': 0}, 
			{'name': 'Top Sites', 
					'values': {'bytes out': 4900070621184.0, 
						'connections':351976704.0, 
						'bytes in':21536893304832.0},
					'flags': 3, 'type': 0}, 
			{'name': 'StatsByClass', 
					'values':{'bytes out': 4888053415936.0,
						'bytes in': 21408400801792.0},
					'flags': 3, 'type': 0}]

		@type start: C{str} or datetime.date
		@param start: Start date in the format YYYY-MM-DD

		@type end: C{str} or datetime.date
		@param end: End date in the format YYYY-MM-DD

		@type path: C{str}
		@param path: The statistics path to list i.e. I{/}

		@type pathtype: C{int}
		@param pathtype: Value path type. For use in firmwares v12.0 and newer.

		@rtype: C{list} of C{dict}
		@returns: A dictionary for each stats node inside a list.

		@raises PLDBUnsupportedInFirmware: If pathtype argument is used withoutfirmware support.

		@attention: pathtype parameter requires v12.0 firmware or newer.
		"""
		
		# We don't use isinstance since that would require importing datetime, that might no be present.
		if start.__class__.__name__ == 'date':
			start = str(start)
			
		if end.__class__.__name__ == 'date':
			end = str(end)

		if pathtype:
			raise PLDBUnsupportedInFirmware("Usage of 'pathtype' is not supported in this firmware version")

		if pathtype == _NONEXISTANT_VALUETYPE:
			raise PLDBUnsupportedInFirmware("Specified 'pathtype' is not supported in this firmware version")
			
			
		cid = self.session.stat_list(start, end, path)
		return plutil._do(self, cid, plutil._SimpleListCB(self, self.session.stat_list_handle, Priority))
	
	def toplist(self, start, end, path, max, pathtype=0):
		"""
		List top N nodes between two dates

		This function exists to limit the amount of data sent between 
		the server and the client.It returns any node which is in the 
		top N list for each of the fields being stored. 
		
		Also if the requested Statistics node contains sub-nodes of more
		than one type then a Top N list would be sent for each type for
		each field.

			>>> res=s.toplist("2009-01-14", "2009-01-14", "/v11 stats", 5, s.VALUETYPE_ROOT)
			>>> for node in res["Toplist, bytes in"]:
			...	if node['type'] == s.VALUETYPE_SERVICE:
			...     	print node
			... 
			{'values': {'bytes out': 30924881920.0, 'bytes in': 6812166656.0}, 'type': 518, 'name': 'BitTorrent transfer'}
			{'values': {'bytes out': 10508003328.0, 'bytes in': 2729902848.0}, 'type': 518, 'name': 'BitTorrent encrypted transfer'}
			{'values': {'bytes out': 2935312384.0, 'bytes in': 2098169216.0}, 'type': 518, 'name': 'HTTP'}
			{'values': {'bytes out': 577248896.0, 'bytes in': 1004788544.0}, 'type': 518, 'name': 'SMTP'}
			{'values': {'bytes out': 7716958720.0, 'bytes in': 923890496.0}, 'type': 518, 'name': 'CIFS'}


		@type start: C{str} or datetime.date
		@param start: Start date in the format YYYY-MM-DD

		@type end: C{str} or datetime.date
		@param end: End date in the format YYYY-MM-DD

		@type path: C{str}
		@param path: The statistics path to list i.e. I{/NetObjects}

		@type max: C{int}
		@param max: Maximum number of nodes to output. If this is zero, all nodes will be returned.

		@type pathtype: C{int}
		@param pathtype: Value path type. For use in firmwares v12.0 and newer.

		@returns: Dict with the top list for each field stored ('In Bytes', 'Out Bytes', 
			'connections', 'Unest. Connections', 'In Connections',
			'Out Connections', 'In Quality of Experience' 'Out Quality of Experience'.
			Each list of is a list of dictonaries for each node.
		
		@rtype: C{dict} of C{list} of C{dict}

		@raises PLDBUnsupportedInFirmware: If pathtype argument is used withoutfirmware support.

		@attention: pathtype parameter requires v12.0 firmware or newer.
		"""
		
		# We don't use isinstance since that would require importing datetime, that might no be present.
		if start.__class__.__name__ == 'date':
			start = str(start)
			
		if end.__class__.__name__ == 'date':
			end = str(end)
			
		if pathtype:
			raise PLDBUnsupportedInFirmware("Usage of 'pathtype' is not supported in this firmware version")

		if pathtype == _NONEXISTANT_VALUETYPE:
			raise PLDBUnsupportedInFirmware("Specified 'pathtype' is not supported in this firmware version")

		cid = self.session.top_list(start, end, path, max)
		nodelist = plutil._do(self, cid, plutil._SimpleListCB(self, self.session.top_list_handle, Priority))

		in_t=[(x['in'],x) for x in nodelist if x.has_key('top_inbound')]
		out_t=[(x['out'],x) for x in nodelist if x.has_key('top_outbound')]
		tot_t=[(x['in'] + x['out'],x) for x in nodelist if x.has_key('top_total')]

		in_t.sort()
		out_t.sort()
		tot_t.sort()
		
		in_t.reverse()
		out_t.reverse()
		tot_t.reverse()
		
		inbound = [x[1] for x in in_t]
		outbound = [x[1] for x in out_t]
		total = [x[1] for x in tot_t]
		
		return {"inbound":inbound, "outbound":outbound, "total":total}

	def graph(self, start, end, path, numvals=24, pathtype=0, mode=1):
		"""
		Retrieve graph data between two dates.

			>>> s.graph("2009-01-05", "2008-01-05", "/Host per services", 24, 0, 1)
			[{'bytes out': 528960416.0, 'connections': 7485.5, 'bytes in': 95168704.0},
			 {'bytes out': 503229952.0, 'connections': 7073.0, 'bytes in': 136215120.0},
			 {'bytes out': 561612864.0, 'connections': 7457.75, 'bytes in': 165719856.0},
			 {'bytes out': 574073536.0, 'connections': 7316.66650390625, 'bytes in': 125461376.0}, 
			 {'bytes out': 575875264.0, 'connections': 7071.0, 'bytes in': 142020688.0}, 
			 {'bytes out': 533374208.0, 'connections': 7379.75, 'bytes in': 128362920.0}, 
			 {'bytes out': 542574656.0, 'connections': 7120.75, 'bytes in': 145863184.0}, 
			 {'bytes out': 627854016.0, 'connections': 7328.91650390625, 'bytes in': 124391448.0},
			 {'bytes out': 356714592.0, 'connections': 8949.1669921875, 'bytes in': 111401304.0}, 
			 {'bytes out': 276146240.0, 'connections': 9937.1669921875, 'bytes in': 121161912.0}, 
			 {'bytes out': 271035360.0, 'connections': 10208.8330078125, 'bytes in': 105619560.0}, 
			 {'bytes out': 282889088.0, 'connections': 9836.9169921875, 'bytes in': 105585656.0}, 
			 {'bytes out': 257417728.0, 'connections': 9915.75, 'bytes in': 118984736.0},
			 {'bytes out': 291532384.0, 'connections': 10910.6669921875, 'bytes in': 198847856.0}, 
			 {'bytes out': 281107264.0, 'connections': 10963.0830078125, 'bytes in': 96303560.0}, 
			 {'bytes out': 261479232.0, 'connections': 10322.6669921875, 'bytes in': 106611800.0}, 
			 {'bytes out': 365030240.0, 'connections': 10602.9169921875, 'bytes in': 233681280.0}, 
			 {'bytes out': 514466784.0, 'connections': 10718.75, 'bytes in': 129939368.0},
			 {'bytes out': 394594048.0, 'connections': 10602.5, 'bytes in': 128155672.0},
			 {'bytes out': 1116160000.0, 'connections': 11515.25, 'bytes in': 149660416.0},
			 {'bytes out': 423695584.0, 'connections': 10865.25, 'bytes in': 121420984.0},
			 {'bytes out': 592000384.0, 'connections': 10532.0, 'bytes in': 228060944.0},
			 {'bytes out': 623388928.0, 'connections': 9260.9169921875, 'bytes in': 198206720.0}, 
			 {'bytes out': 323936608.0, 'connections': 8981.25, 'bytes in': 117023752.0}]	

		@type start: C{str} or datetime.date
		@param start: Start date in the format YYYY-MM-DD

		@type end: C{str} or datetime.date
		@param end: End date in the format YYYY-MM-DD

		@type path: C{str}
		@param path: The statistics path to list i.e. I{/NetObjects}

		@type numvals: C{int}
		@param numvals: the number of values that should be extracted.

		@type pathtype: C{int}
		@param pathtype: Value path type.  

		@type mode: C{int}
		@param mode: throughput rate or actual byte/connection count.

		@rtype: C{dict} {list} of C{dict}
		@returns: Dictionaries with data.

		@raises PLDBUnsupportedInFirmware: If pathtype/mode arguments are used without firmware support.

		@attention: pathtype parameter requires v12.0 firmware or newer.
		@attention: mode parameter requires v12.0 firmware or newer.
		"""

		# We don't use isinstance since that would require importing datetime, that might no be present.
		if start.__class__.__name__ == 'date':
			start = str(start)
			
		if end.__class__.__name__ == 'date':
			end = str(end)
			
		if pathtype:
			raise PLDBUnsupportedInFirmware("Usage of 'pathtype' is not supported in this firmware version")

		if pathtype == _NONEXISTANT_VALUETYPE:
			raise PLDBUnsupportedInFirmware("Specified 'pathtype' is not supported in this firmware version")

		if mode != 1:
			raise PLDBUnsupportedInFirmware("Usage of 'mode' is not supported in this firmware version")

		if mode == _NONEXISTANT_GRAPH_MODE:
			raise PLDBUnsupportedInFirmware("Specified 'mode' is not supported in this firmware version")


		cid = self.session.stat_graph(start, end, path, numvals)
		return plutil._do(self, cid, plutil._ListCB(self, self.session.stat_graph_handle, PLDB.PLGraphList, False))

	def path_is_expandable(self, start, end, path, pathtype=0):
		"""
		Verify that a path is expandable.

		Using datetime.date is new since version 11.3beta5.
		
		@type start: C{str} or datetime.date
		@param start: Start date in the format YYYY-MM-DD
		
		@type end: C{str} or datetime.date
		@param end: End date in the format YYYY-MM-DD

		@type path: C{str}
		@param path: The statistics path to list i.e. I{/NetObjects}

		@type pathtype: C{int}
		@param pathtype: Value path type.  

		@returns: True if path is expandable.
		@rtype: C{boolean}

		@raises PLDBUnsupportedInFirmware: If pathtype argument is used without firmware support.

		@attention: pathtype parameter requires v12.0 firmware or newer.
		"""
		
		# We don't use isinstance since that would require importing datetime, that might no be present.
		if start.__class__.__name__ == 'date':
			start = str(start)
			
		if end.__class__.__name__ == 'date':
			end = str(end)
			
		if pathtype:
			raise PLDBUnsupportedInFirmware("Usage of 'pathtype' is not supported in this firmware version")

		if pathtype == _NONEXISTANT_VALUETYPE:
			raise PLDBUnsupportedInFirmware("Specified 'pathtype' is not supported in this firmware version")

		cid = self.session.stat_is_expandable (start, end, path)
		if plutil._do(self, cid, plutil._AddCB(self, self.session.stat_is_expandable_handle)):
			return True
		return False

	def transfer_get(self, start, end, path, pathtype=0):
		"""
		Retrieves total transfer for path between two dates.

			>>> r.transfer_get("2006-01-20", "2006-01-21", "/NetObjects")
			{'priority': 129, 'out': 1458559580273L, 'name': '/NetObjects', 'in': 1632574443658L}

		@type start: C{str} or datetime.date
		@param start: Start date in the format YYYY-MM-DD

		@type end: C{str} or datetime.date
		@param end: End date in the format YYYY-MM-DD
		
		@type path: C{str}
		@param path: The statistics path to list i.e. I{/NetObjects}

		@type pathtype: C{int}
		@param pathtype: Value path type.  

		@returns: Dictionary containing transfer information.
		@rtype: C{dict}

		@raises RuntimeError: B{"Could not retrieve data for path"}:
			If statistic path could not be read.
		@raises PLDBUnsupportedInFirmware: If pathtype argument is used without firmware support.

		@attention: pathtype parameter requires v12.0 firmware or newer.
		"""
		
		# We don't use isinstance since that would require importing datetime, that might no be present.
		if start.__class__.__name__ == 'date':
			start = str(start)
			
		if end.__class__.__name__ == 'date':
			end = str(end)
			
		if pathtype:
			raise PLDBUnsupportedInFirmware("Usage of 'pathtype' is not supported in this firmware version")

		if pathtype == _NONEXISTANT_VALUETYPE:
			raise PLDBUnsupportedInFirmware("Specified 'pathtype' is not supported in this firmware version")

		cid = self.session.stat_transfer_get (start, end, path)
		res = plutil._do(self, cid, plutil._AddCB(self, self.session.stat_transfer_get_handle))
		res["priority"] = Priority(res["priority"])
		return res
