"""
This module contains MPLSObject implementation.

There is no need to import this module directly. Use the methods of
the L{Ruleset object <ruleset.Ruleset>} to create and manipulate
MPLSObjects.

    >>> o = r.object_add("/MPLSObjects/Test")
    >>> o.add("100000")
    >>> o.add("200000-210000")
    >>> print o.items
    ['100000', '200000-210000']

@var __revision__: Revision identifier.
"""

__revision__ = '%K%'

import plobject
from thisversion import PLDB
import plutil

def _verify_mpls(mpls_id):
	if mpls_id < 0 or mpls_id > (1 << 20):
		return False
	return True

class MPLSObject(plobject.PLObject):
	"""
	MPLSObject Implementation

	Object path: /MPLSObjects

	@cvar Item: MPLSObject Item implementation.
	"""

	__slots__ = ('_prop', '_pldb', '_callback', '_item_added')
	_may_exclude = True

	base_path = plutil.roConst("/MPLSObjects", "Only the type portion of the path")
	type = plutil.roConst(PLDB.OBJECT_TYPE_MPLSOBJECT, "Number representing the type of object")

	def _create_item(self, value):
		if isinstance(value, basestring) and "-" in value:
			first, last = value.split("-", 1)
			try:
				item = self.Item(self, value1=first, value2=last, type=PLDB.MPLSOBJECT_ITEM_TYPE_RANGE)
			except:
				raise ValueError("Format is not correct for this item type")
				
			if item._prop.value2 < item._prop.value1:
				raise ValueError("First label should not be smaller than the last label in a range.")
		else:
			try:
				item = self.Item(self, value1=value, value2=0, type=PLDB.MPLSOBJECT_ITEM_TYPE_LABEL)
			except:
				raise ValueError("Format is not correct for this item type")

		return item

	def add(self, value, exclude=False):
		"""
		Adds a MPLSObject item to this MPLSObject.

		@type value: C{str}
		@param value: The value is a string describing the MPLSObject item. The type of the item
			will be automatically detected from the string. The following syntax is valid:
				- Single label, just give the plain MPLS id I{nn}
				- Label range, give the start and stop MPLS id with a dash as delimiter I{nn-mm}

		@type exclude: C{boolean}
		@param exclude: If this is set to True the item will be a exclude
			Item instead of a include Item.

		"""
		plobject.PLObject.add(self, value, exclude)


	class Item(plobject.PLObject.Item):
		"""
		MPLSObject Item implementation.
		Should not be instantiated, use L{MPLSObject.add} instead.

		@type exclude: bool
		@type type: int
		@type value1: int
		@type value2: int
		"""
		__slots__ = ('_prop', '_parent')
		_Prop = plutil.Prop(exclude=bool, type=int, value1=int, value2=int, *plobject._item_props)

		def __str__(self):
			if self._prop.exclude:
				ex = "!"
			else:
				ex = ""

			if self.type == PLDB.MPLSOBJECT_ITEM_TYPE_LABEL:
				return "%s%d" % (ex, self._prop.value1)
			elif self.type == PLDB.MPLSOBJECT_ITEM_TYPE_RANGE:
				return "%s%d-%d" % (ex, self._prop.value1, self._prop.value2)
			else:
				raise ValueError("Wrong item-type for this object")
			
		type = property(lambda self: self._prop.type, doc="Internal number representing the item type (single label, or range)")
		exclude = property(lambda self: self._prop.exclude, doc="True = exclude, False = include")
		value1 = property(lambda self: self._prop.value1, doc="label (or first label in range)")
		value2 = property(lambda self: self._prop.value2, doc="last label in the label range (0 for single label)")
